// ----------------------------------------------------------------------------------------
// Name        : Test.cpp
// Description : Test the RectPlacement class
//               (C) Copyright 2000-2002 by Javier Arevalo
//               This code is free to use and modify for all purposes
// ----------------------------------------------------------------------------------------

#include <stdio.h>
#include <algorithm>

#include "RectPlacement.h"

#define MAX_TEXTURE_W 256
#define MAX_TEXTURE_H 256

#define NUM_SUBRECTS  500
#define MIN_SUBRECT    15
#define MAX_SUBRECT    50

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------

struct TSubRect: public CRectPlacement::TRect
{
  int n;      // Original index of this subrect, before sorting
  int nTex;   // Texture in which this subrect will be placed.

  TSubRect() { }
  TSubRect(int _w, int _h, int _n): TRect(0, 0, _w, _h), n(_n), nTex(0) { }
};

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------

typedef std::vector<TSubRect>       CSubRectArray;
typedef std::vector<CRectPlacement> CTextureArray;

// --------------------------------------------------------------------------------
// Name        : CreateLetters
// Description : Here's where you calculate your rectangles,
// --------------------------------------------------------------------------------
void CreateLetters  (CSubRectArray &vecSubRects)
{
  vecSubRects.clear();

  // Let's just fill this with random stuff
  for (int i = 0; i < NUM_SUBRECTS; i++)
    vecSubRects.push_back(TSubRect((rand() % MAX_SUBRECT) + MIN_SUBRECT, (rand() % MAX_SUBRECT) + MIN_SUBRECT, i));
}

// --------------------------------------------------------------------------------
// Name        : CreateTextures
// Description : Create array of textures containing all subrects
// --------------------------------------------------------------------------------
void CreateTextures (CTextureArray &vecTextures, CSubRectArray &vecSubRects, int maxTexW, int maxTexH)
{
  // Sort the subRects based on dimensions, larger dimension goes first.
  std::sort(vecSubRects.begin(), vecSubRects.end(), CRectPlacement::TRect::Greater);

  // Generate the first texture
  vecTextures.clear();
  vecTextures.push_back(CRectPlacement());

  // Add all subrects
  for (CSubRectArray::iterator it = vecSubRects.begin();
       it != vecSubRects.end();
       ++it)
  {
    // We make sure we leave one pixel between subrects, so texels don't bleed with bilinear.
    CRectPlacement::TRect r(0, 0, it->w+1, it->h+1);

    // If an existing texture has actual space 
    bool bPlaced = false;
    for (int i = 0; !bPlaced && i < vecTextures.size(); i++)
    {
      bPlaced = vecTextures[i].AddAtEmptySpotAutoGrow(&r, maxTexW, maxTexH);
      if (bPlaced)
        it->nTex = i;
    }

    // Try starting a new texture and fit the rect in there
    if (!bPlaced)
    {
      vecTextures.push_back(CRectPlacement());
      bPlaced = vecTextures[vecTextures.size()-1].AddAtEmptySpotAutoGrow(&r, maxTexW, maxTexH);
      if (bPlaced)
        it->nTex = vecTextures.size()-1;
      else
        printf("ERROR: Subrect is too big to fit in texture!", it->w, it->h);
    }

    // If correctly placed in a texture, the coords are returned in r.x and r.y
    // Store them.
    if (bPlaced)
    {
      it->x = r.x;
      it->y = r.y;
    }
  }
}

// --------------------------------------------------------------------------------
// Name        : main
// Description : heh
// --------------------------------------------------------------------------------
int main()
{
  CTextureArray vecTextures;
  CSubRectArray vecSubRects;

  CreateLetters(vecSubRects);
  CreateTextures(vecTextures, vecSubRects, MAX_TEXTURE_W, MAX_TEXTURE_H);

  {
    for (CSubRectArray::const_iterator it = vecSubRects.begin();
         it != vecSubRects.end();
         ++it)
    {
      printf("Subrect %d (originally %d), size %dx%d, goes into texture %d at pos %d,%d\n",
             it - vecSubRects.begin(), it->n, it->w, it->h, it->nTex, it->x, it->y);
    }
  }

  printf("Created a total of %d textures\n", vecTextures.size());
  {
    for (CTextureArray::const_iterator it = vecTextures.begin();
         it != vecTextures.end();
         ++it)
    {
      printf("  Texture %d, size %dx%d, Coverage %d / %d (%d%%)\n",
             it - vecTextures.begin(), it->GetW(), it->GetH(),
             it->GetArea(), it->GetTotalArea(), it->GetArea()*100/it->GetTotalArea());
    }
  }
  return 0;
}
